# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from particle.pdgid import (
    A,
    J,
    L,
    S,
    Z,
    charge,
    has_bottom,
    has_charm,
    has_down,
    has_fundamental_anti,
    has_strange,
    has_top,
    has_up,
    is_baryon,
    is_diquark,
    is_dyon,
    is_excited_quark_or_lepton,
    is_gauge_boson_or_higgs,
    is_generator_specific,
    is_hadron,
    is_lepton,
    is_meson,
    is_nucleus,
    is_pentaquark,
    is_Qball,
    is_quark,
    is_Rhadron,
    is_sm_gauge_boson_or_higgs,
    is_sm_lepton,
    is_sm_quark,
    is_special_particle,
    is_SUSY,
    is_technicolor,
    is_valid,
    j_spin,
    l_spin,
    s_spin,
    three_charge,
)


def test_charge(PDGIDs):
    assert charge(PDGIDs.Photon) == 0
    assert charge(PDGIDs.Gluon) == 0
    assert charge(PDGIDs.Electron) == -1
    assert charge(PDGIDs.AntiMuon) == +1
    assert charge(PDGIDs.jpsi) == 0
    assert charge(PDGIDs.Upsilon_1S) == 0
    assert charge(PDGIDs.PiPlus) == +1
    assert charge(PDGIDs.KMinus) == -1
    assert charge(PDGIDs.D0) == 0
    assert charge(PDGIDs.DPlus) == +1
    assert charge(PDGIDs.DsPlus) == +1
    assert charge(PDGIDs.B0) == 0
    assert charge(PDGIDs.Bs) == 0
    assert charge(PDGIDs.BcPlus) == +1
    assert charge(PDGIDs.Proton) == +1
    assert charge(PDGIDs.LcPlus) == +1
    assert charge(PDGIDs.Lb) == 0
    assert charge(PDGIDs.DD1) == -2 / 3
    assert charge(PDGIDs.SD0) == -2 / 3
    assert charge(PDGIDs.Invalid1) is None
    assert charge(PDGIDs.Invalid2) is None


def test_three_charge(PDGIDs):
    assert three_charge(PDGIDs.Photon) == 0
    assert three_charge(PDGIDs.Electron) == -3
    assert three_charge(PDGIDs.jpsi) == 0
    assert three_charge(PDGIDs.Upsilon_1S) == 0
    assert three_charge(PDGIDs.KMinus) == -3
    assert three_charge(PDGIDs.D0) == 0
    assert three_charge(PDGIDs.Proton) == +3
    assert three_charge(PDGIDs.LcPlus) == +3
    assert three_charge(PDGIDs.Lb) == 0
    assert three_charge(PDGIDs.DD1) == -2
    assert three_charge(PDGIDs.SD0) == -2
    assert three_charge(PDGIDs.Invalid1) is None
    assert three_charge(PDGIDs.Invalid2) is None


def test_is_valid(PDGIDs):
    assert is_valid(PDGIDs.Photon)
    assert is_valid(PDGIDs.Gluon)
    assert is_valid(PDGIDs.Electron)
    assert is_valid(PDGIDs.AntiMuon)
    assert is_valid(PDGIDs.jpsi)
    assert is_valid(PDGIDs.Upsilon_1S)
    assert is_valid(PDGIDs.PiPlus)
    assert is_valid(PDGIDs.KMinus)
    assert is_valid(PDGIDs.D0)
    assert is_valid(PDGIDs.DPlus)
    assert is_valid(PDGIDs.DsPlus)
    assert is_valid(PDGIDs.B0)
    assert is_valid(PDGIDs.Bs)
    assert is_valid(PDGIDs.BcPlus)
    assert is_valid(PDGIDs.Proton)
    assert is_valid(PDGIDs.LcPlus)
    assert is_valid(PDGIDs.Lb)
    assert is_valid(PDGIDs.DD1)
    assert is_valid(PDGIDs.SD0)
    assert not is_valid(PDGIDs.Invalid1)
    assert not is_valid(PDGIDs.Invalid2)


def test_is_quark(PDGIDs):
    _quarks = (
        PDGIDs.DQuark,
        PDGIDs.UQuark,
        PDGIDs.SQuark,
        PDGIDs.CQuark,
        PDGIDs.BQuark,
        PDGIDs.TQuark,
        PDGIDs.BPrimeQuark,
        PDGIDs.TPrimeQuark,
    )
    _non_quarks = [id for id in PDGIDs if id not in _quarks]
    for id in _quarks:
        assert is_quark(id)
    for id in _non_quarks:
        assert not is_quark(id)


def test_is_sm_quark(PDGIDs):
    _sm_quarks = (
        PDGIDs.DQuark,
        PDGIDs.UQuark,
        PDGIDs.SQuark,
        PDGIDs.CQuark,
        PDGIDs.BQuark,
        PDGIDs.TQuark,
    )
    _non_sm_quarks = [id for id in PDGIDs if id not in _sm_quarks]
    for id in _sm_quarks:
        assert is_sm_quark(id)
    for id in _non_sm_quarks:
        assert not is_sm_quark(id)


def test_is_lepton(PDGIDs):
    _leptons = (
        PDGIDs.Electron,
        PDGIDs.Positron,
        PDGIDs.Muon,
        PDGIDs.AntiMuon,
        PDGIDs.Tau,
        PDGIDs.TauPrime,
        PDGIDs.Nu_e,
        PDGIDs.NuBar_tau,
    )
    _non_leptons = [id for id in PDGIDs if id not in _leptons]
    for id in _leptons:
        assert is_lepton(id)
    for id in _non_leptons:
        assert not is_lepton(id)


def test_is_sm_lepton(PDGIDs):
    _sm_leptons = (
        PDGIDs.Electron,
        PDGIDs.Positron,
        PDGIDs.Muon,
        PDGIDs.AntiMuon,
        PDGIDs.Tau,
        PDGIDs.Nu_e,
        PDGIDs.NuBar_tau,
    )
    _non_sm_leptons = [id for id in PDGIDs if id not in _sm_leptons]
    for id in _sm_leptons:
        assert is_sm_lepton(id)
    for id in _non_sm_leptons:
        assert not is_sm_lepton(id)


def test_is_meson(PDGIDs):
    _mesons = (
        PDGIDs.jpsi,
        PDGIDs.psi_2S,
        PDGIDs.Upsilon_1S,
        PDGIDs.Upsilon_4S,
        PDGIDs.Pi0,
        PDGIDs.PiPlus,
        PDGIDs.eta,
        PDGIDs.eta_prime,
        PDGIDs.a_0_1450_plus,
        PDGIDs.KL,
        PDGIDs.KS,
        PDGIDs.KMinus,
        PDGIDs.phi,
        PDGIDs.omega,
        PDGIDs.rho_770_minus,
        PDGIDs.K1_1270_0,
        PDGIDs.K1_1400_0,
        PDGIDs.rho_1700_0,
        PDGIDs.a2_1320_minus,
        PDGIDs.omega_3_1670,
        PDGIDs.f_4_2300,
        PDGIDs.D0,
        PDGIDs.DPlus,
        PDGIDs.DsPlus,
        PDGIDs.B0,
        PDGIDs.BPlus,
        PDGIDs.Bs,
        PDGIDs.BcPlus,
        PDGIDs.Pi0TC,
        PDGIDs.PiMinusTC,
        PDGIDs.T0,
        PDGIDs.Reggeon,
        PDGIDs.Pomeron,
        PDGIDs.Odderon,
        PDGIDs.RPlus_TTildeDbar,
        PDGIDs.R0_GTildeG,
    )
    _non_mesons = [id for id in PDGIDs if id not in _mesons]
    for id in _mesons:
        assert is_meson(id)
    for id in _non_mesons:
        assert not is_meson(id)


def test_is_meson_B_mass_eigenstates():
    # Test special IDs of B(L)0, B(sL)0, B(H)0, B(sH)0
    for pdgid in {150, 350, 510, 530}:
        assert is_meson(pdgid)


def test_is_baryon(PDGIDs):
    _baryons = (
        PDGIDs.Proton,
        PDGIDs.AntiNeutron,
        PDGIDs.HydrogenNucleus,
        PDGIDs.Lambda,
        PDGIDs.Sigma0,
        PDGIDs.SigmaPlus,
        PDGIDs.SigmaMinus,
        PDGIDs.Xi0,
        PDGIDs.AntiXiMinus,
        PDGIDs.OmegaMinus,
        PDGIDs.LcPlus,
        PDGIDs.Lb,
        PDGIDs.LtPlus,
        PDGIDs.RPlusPlus_GTildeUUU,
        PDGIDs.UCbarCUDPentaquark,
        PDGIDs.AntiUCbarCUDPentaquark,
    )
    _non_baryons = [id for id in PDGIDs if id not in _baryons]
    for id in _baryons:
        assert is_baryon(id)
    for id in _non_baryons:
        assert not is_baryon(id)


def test_is_baryon_old_codes_diffractive():
    # Test old codes for diffractive p and n (MC usage)
    assert is_baryon(2110)
    assert is_baryon(2210)


def test_is_hadron(PDGIDs):
    for id in PDGIDs:
        assert is_hadron(id) == (is_meson(id) or is_baryon(id))


def test_is_pentaquark(PDGIDs):
    _pentaquarks = (PDGIDs.UCbarCUDPentaquark, PDGIDs.AntiUCbarCUDPentaquark)
    _non_pentaquarks = [id for id in PDGIDs if id not in _pentaquarks]
    assert is_pentaquark(PDGIDs.UCbarCUDPentaquark)
    assert is_pentaquark(PDGIDs.AntiUCbarCUDPentaquark)
    for id in _non_pentaquarks:
        assert not is_pentaquark(id)


def test_is_gauge_boson_or_higgs(PDGIDs):
    _gb_and_higgs = (
        PDGIDs.Gluon,
        PDGIDs.Photon,
        PDGIDs.Z0,
        PDGIDs.WMinus,
        PDGIDs.HiggsBoson,
        PDGIDs.ZPrime,
        PDGIDs.Graviton,
    )
    _non_gb_and_higgs = [id for id in PDGIDs if id not in _gb_and_higgs]
    for id in _gb_and_higgs:
        assert is_gauge_boson_or_higgs(id)
    for id in _non_gb_and_higgs:
        assert not is_gauge_boson_or_higgs(id)


def test_is_sm_gauge_boson_or_higgs(PDGIDs):
    _sm_gb_and_higgs = (
        PDGIDs.Gluon,
        PDGIDs.Photon,
        PDGIDs.Z0,
        PDGIDs.WMinus,
        PDGIDs.HiggsBoson,
    )
    _non_sm_gb_and_higgs = [id for id in PDGIDs if id not in _sm_gb_and_higgs]
    for id in _sm_gb_and_higgs:
        assert is_sm_gauge_boson_or_higgs(id)
    for id in _non_sm_gb_and_higgs:
        assert not is_sm_gauge_boson_or_higgs(id)


def test_is_generator_specific(PDGIDs):
    _generator_specific = (PDGIDs.AntiCHadron,)
    _non_generator_specific = [id for id in PDGIDs if id not in _generator_specific]
    for id in _generator_specific:
        assert is_generator_specific(id)
    for id in _non_generator_specific:
        assert not is_generator_specific(id)


def test_is_special_particle(PDGIDs):
    _special_particle = (
        PDGIDs.Graviton,
        PDGIDs.Reggeon,
        PDGIDs.Pomeron,
        PDGIDs.Odderon,
        PDGIDs.AntiCHadron,
    )
    _non_special_particle = [id for id in PDGIDs if id not in _special_particle]
    for id in _special_particle:
        assert is_special_particle(id)
    for id in _non_special_particle:
        assert not is_special_particle(id)


def test_is_nucleus(PDGIDs):
    _nuclei = (
        PDGIDs.Proton,
        PDGIDs.AntiNeutron,
        PDGIDs.HydrogenNucleus,
        PDGIDs.Carbon12,
    )
    _non_nuclei = [id for id in PDGIDs if id not in _nuclei]
    for id in _nuclei:
        assert is_nucleus(id)
    for id in _non_nuclei:
        assert not is_nucleus(id)
    # test 10-digit IDs that does not conform with form for nuclei (should start with +/- 10)
    assert not is_nucleus(2000000010)
    assert not is_nucleus(1100000010)


def test_is_diquark(PDGIDs):
    _diquarks = (PDGIDs.DD1, PDGIDs.SD0)
    _non_diquarks = [id for id in PDGIDs if id not in _diquarks]
    for id in _diquarks:
        assert is_diquark(id)
    for id in _non_diquarks:
        assert not is_diquark(id)


def test_is_Rhadron(PDGIDs):
    _Rhadrons = (PDGIDs.RPlus_TTildeDbar, PDGIDs.R0_GTildeG, PDGIDs.RPlusPlus_GTildeUUU)
    _non_Rhadrons = [id for id in PDGIDs if id not in _Rhadrons]
    for id in _Rhadrons:
        assert is_Rhadron(id)
    for id in _non_Rhadrons:
        assert not is_Rhadron(id)


def test_is_Qball(PDGIDs):
    _Qballs = (PDGIDs.QBall1, PDGIDs.QBall2)
    _non_Qballs = [id for id in PDGIDs if id not in _Qballs]
    for id in _Qballs:
        assert is_Qball(id)
    for id in _non_Qballs:
        assert not is_Qball(id)


def test_is_dyon(PDGIDs):
    _dyons = (PDGIDs.DyonSameMagElecChargeSign, PDGIDs.DyonOppositeMagElecChargeSign)
    _non_dyons = [id for id in PDGIDs if id not in _dyons]
    for id in _dyons:
        assert is_dyon(id)
    for id in _non_dyons:
        assert not is_dyon(id)


def test_is_SUSY(PDGIDs):
    _susy = (PDGIDs.Gluino, PDGIDs.Gravitino, PDGIDs.STildeL, PDGIDs.CTildeR)
    _non_susy = [id for id in PDGIDs if id not in _susy]
    for id in _susy:
        assert is_SUSY(id)
    for id in _non_susy:
        assert not is_SUSY(id)


def test_is_technicolor(PDGIDs):
    _technicolor = (PDGIDs.Pi0TC, PDGIDs.PiMinusTC)
    _non_technicolor = [id for id in PDGIDs if id not in _technicolor]
    for id in _technicolor:
        assert is_technicolor(id)
    for id in _non_technicolor:
        assert not is_technicolor(id)


def test_is_excited_quark_or_lepton(PDGIDs):
    _excited_quark_or_lepton = (PDGIDs.UQuarkStar, PDGIDs.AntiElectronStar)
    _non_excited_quark_or_lepton = [
        id for id in PDGIDs if id not in _excited_quark_or_lepton
    ]
    for id in _excited_quark_or_lepton:
        assert is_excited_quark_or_lepton(id)
    for id in _non_excited_quark_or_lepton:
        assert not is_excited_quark_or_lepton(id)


def test_has_down(PDGIDs):
    assert not has_down(PDGIDs.Photon)
    assert not has_down(PDGIDs.Gluon)
    assert not has_down(PDGIDs.Electron)
    assert not has_down(PDGIDs.AntiMuon)
    assert not has_down(PDGIDs.jpsi)
    assert not has_down(PDGIDs.Upsilon_1S)
    assert has_down(PDGIDs.PiPlus)
    assert not has_down(PDGIDs.KMinus)
    assert not has_down(PDGIDs.D0)
    assert has_down(PDGIDs.DPlus)
    assert not has_down(PDGIDs.DsPlus)
    assert has_down(PDGIDs.B0)
    assert not has_down(PDGIDs.Bs)
    assert not has_down(PDGIDs.BcPlus)
    assert has_down(PDGIDs.Proton)
    assert has_down(PDGIDs.LcPlus)
    assert has_down(PDGIDs.Lb)
    assert has_down(PDGIDs.DD1)
    assert has_down(PDGIDs.SD0)
    assert not has_down(PDGIDs.Invalid1)
    assert not has_down(PDGIDs.Invalid2)


def test_has_up(PDGIDs):
    assert not has_up(PDGIDs.Photon)
    assert not has_up(PDGIDs.Gluon)
    assert not has_up(PDGIDs.Electron)
    assert not has_up(PDGIDs.AntiMuon)
    assert not has_up(PDGIDs.jpsi)
    assert not has_up(PDGIDs.Upsilon_1S)
    assert has_up(PDGIDs.PiPlus)
    assert has_up(PDGIDs.KMinus)
    assert has_up(PDGIDs.D0)
    assert not has_up(PDGIDs.DPlus)
    assert not has_up(PDGIDs.DsPlus)
    assert not has_up(PDGIDs.B0)
    assert not has_up(PDGIDs.Bs)
    assert not has_up(PDGIDs.BcPlus)
    assert has_up(PDGIDs.Proton)
    assert has_up(PDGIDs.LcPlus)
    assert has_up(PDGIDs.Lb)
    assert not has_up(PDGIDs.DD1)
    assert not has_up(PDGIDs.SD0)
    assert not has_up(PDGIDs.Invalid1)
    assert not has_up(PDGIDs.Invalid2)


def test_has_strange(PDGIDs):
    assert not has_strange(PDGIDs.Photon)
    assert not has_strange(PDGIDs.Gluon)
    assert not has_strange(PDGIDs.Electron)
    assert not has_strange(PDGIDs.AntiMuon)
    assert not has_strange(PDGIDs.jpsi)
    assert not has_strange(PDGIDs.Upsilon_1S)
    assert not has_strange(PDGIDs.PiPlus)
    assert has_strange(PDGIDs.KMinus)
    assert not has_strange(PDGIDs.D0)
    assert not has_strange(PDGIDs.DPlus)
    assert has_strange(PDGIDs.DsPlus)
    assert not has_strange(PDGIDs.B0)
    assert has_strange(PDGIDs.Bs)
    assert not has_strange(PDGIDs.BcPlus)
    assert not has_strange(PDGIDs.Proton)
    assert not has_strange(PDGIDs.LcPlus)
    assert not has_strange(PDGIDs.Lb)
    assert not has_strange(PDGIDs.DD1)
    assert has_strange(PDGIDs.SD0)
    assert not has_strange(PDGIDs.Invalid1)
    assert not has_strange(PDGIDs.Invalid2)


def test_has_charm(PDGIDs):
    _with_charm_content = (
        PDGIDs.jpsi,
        PDGIDs.psi_2S,
        PDGIDs.D0,
        PDGIDs.DPlus,
        PDGIDs.DsPlus,
        PDGIDs.BcPlus,
        PDGIDs.LcPlus,
        PDGIDs.UCbarCUDPentaquark,
        PDGIDs.AntiUCbarCUDPentaquark,
    )
    _without_charm_content = [id for id in PDGIDs if id not in _with_charm_content]
    for id in _with_charm_content:
        assert has_charm(id)
    for id in _without_charm_content:
        assert not has_charm(id)


def test_has_bottom(PDGIDs):
    _with_bottom_content = (
        PDGIDs.Upsilon_1S,
        PDGIDs.Upsilon_4S,
        PDGIDs.B0,
        PDGIDs.BPlus,
        PDGIDs.Bs,
        PDGIDs.BcPlus,
        PDGIDs.Lb,
    )
    _without_bottom_content = [id for id in PDGIDs if id not in _with_bottom_content]
    for id in _with_bottom_content:
        assert has_bottom(id)
    for id in _without_bottom_content:
        assert not has_bottom(id)


def test_has_top(PDGIDs):
    assert has_top(PDGIDs.T0)
    assert has_top(PDGIDs.LtPlus)
    _no_top = [
        id for id in PDGIDs if id not in (PDGIDs.T0, PDGIDs.LtPlus)
    ]  # top quark should also return has_top(6)==False !
    for id in _no_top:
        assert not has_top(id)


def test_has_fundamental_anti(PDGIDs):
    # Particles that are "fundamental" and not their own antiparticle
    _yep = (
        PDGIDs.WMinus,
        PDGIDs.Electron,
        PDGIDs.Positron,
        PDGIDs.Muon,
        PDGIDs.AntiMuon,
        PDGIDs.Tau,
        PDGIDs.TauPrime,
        PDGIDs.Nu_e,
        PDGIDs.NuBar_tau,
        PDGIDs.DQuark,
        PDGIDs.UQuark,
        PDGIDs.SQuark,
        PDGIDs.CQuark,
        PDGIDs.BQuark,
        PDGIDs.TQuark,
        PDGIDs.BPrimeQuark,
        PDGIDs.TPrimeQuark,
        PDGIDs.UQuarkStar,
        PDGIDs.AntiElectronStar,
        PDGIDs.STildeL,
        PDGIDs.CTildeR,
        PDGIDs.AntiCHadron,
    )
    _nope = [id for id in PDGIDs if id not in _yep]
    for id in _yep:
        assert has_fundamental_anti(id)
    for id in _nope:
        assert not has_fundamental_anti(id)


def mesons_JSL_states_list(PDGIDs, jsl):
    _states = {
        "000": (
            PDGIDs.Pi0,
            PDGIDs.PiPlus,
            PDGIDs.eta,
            PDGIDs.eta_prime,
            PDGIDs.KL,
            PDGIDs.KS,
            PDGIDs.KMinus,
            PDGIDs.D0,
            PDGIDs.DPlus,
            PDGIDs.DsPlus,
            PDGIDs.B0,
            PDGIDs.BPlus,
            PDGIDs.Bs,
            PDGIDs.BcPlus,
            PDGIDs.T0,
        ),
        "011": (PDGIDs.a_0_1450_plus,),
        "101": (PDGIDs.K1_1270_0,),
        "110": (PDGIDs.rho_770_minus,),
        "111": (PDGIDs.K1_1400_0,),
        "112": (PDGIDs.rho_1700_0,),
        # '202': (),
        "211": (PDGIDs.a2_1320_minus,),
        # '212': (),
        # '213': (),
        # '303': (),
        "312": (PDGIDs.omega_3_1670,),
        # '313': (),
        # '314': (),
        # '404': (),
        # '413': (),
        # '414': (),
        # '415': ()
    }
    return _states[jsl]


def test_JSL_mesons(PDGIDs):
    _JSL_eq_000 = mesons_JSL_states_list(PDGIDs, "000")
    _JSL_eq_011 = mesons_JSL_states_list(PDGIDs, "011")
    _JSL_eq_101 = mesons_JSL_states_list(PDGIDs, "101")
    _JSL_eq_110 = mesons_JSL_states_list(PDGIDs, "110")
    _JSL_eq_111 = mesons_JSL_states_list(PDGIDs, "111")
    _JSL_eq_112 = mesons_JSL_states_list(PDGIDs, "112")
    _JSL_eq_211 = mesons_JSL_states_list(PDGIDs, "211")
    _JSL_eq_312 = mesons_JSL_states_list(PDGIDs, "312")

    for id in _JSL_eq_000:
        assert J(id) == 0
        assert S(id) == 0
        assert L(id) == 0
    for id in _JSL_eq_011:
        assert J(id) == 0
        assert S(id) == 1
        assert L(id) == 1
    for id in _JSL_eq_101:
        assert J(id) == 1
        assert S(id) == 0
        assert L(id) == 1
    for id in _JSL_eq_110:
        assert J(id) == 1
        assert S(id) == 1
        assert L(id) == 0
    for id in _JSL_eq_111:
        assert J(id) == 1
        assert S(id) == 1
        assert L(id) == 1
    for id in _JSL_eq_112:
        assert J(id) == 1
        assert S(id) == 1
        assert L(id) == 2
    for id in _JSL_eq_211:
        assert J(id) == 2
        assert S(id) == 1
        assert L(id) == 1
    for id in _JSL_eq_312:
        assert J(id) == 3
        assert S(id) == 1
        assert L(id) == 2


def test_JSL_badly_known_mesons(PDGIDs):
    assert j_spin(PDGIDs.f_4_2300) == 9
    assert s_spin(PDGIDs.f_4_2300) is None
    assert l_spin(PDGIDs.f_4_2300) is None


def test_J_non_mesons(PDGIDs):
    # TODO:  test special particles, supersymmetric particles, R-hadrons, di-quarks, nuclei and pentaquarks
    _J_eq_0 = ()
    _J_eq_1 = (
        PDGIDs.Gluon,
        PDGIDs.Photon,
        PDGIDs.Z0,
        PDGIDs.jpsi,
        PDGIDs.psi_2S,
        PDGIDs.Upsilon_1S,
        PDGIDs.Upsilon_4S,
        PDGIDs.K1_1270_0,
    )
    _J_eq_1over2 = (
        PDGIDs.Electron,
        PDGIDs.Positron,
        PDGIDs.Muon,
        PDGIDs.AntiMuon,
        PDGIDs.Tau,
        PDGIDs.Nu_e,
        PDGIDs.NuBar_tau,
        PDGIDs.DQuark,
        PDGIDs.UQuark,
        PDGIDs.SQuark,
        PDGIDs.CQuark,
        PDGIDs.BQuark,
        PDGIDs.TQuark,
        PDGIDs.Proton,
        PDGIDs.AntiNeutron,
        PDGIDs.Lambda,
        PDGIDs.Sigma0,
        PDGIDs.SigmaPlus,
        PDGIDs.SigmaMinus,
        PDGIDs.Xi0,
        PDGIDs.AntiXiMinus,
        PDGIDs.LcPlus,
        PDGIDs.Lb,
        PDGIDs.LtPlus,
        PDGIDs.STildeL,
        PDGIDs.CTildeR,
    )
    _J_eq_3over2 = (PDGIDs.OmegaMinus,)
    _invalid_pdgids = (PDGIDs.Invalid1, PDGIDs.Invalid2)
    # cases not dealt with in the code, where None is returned
    _J_eq_None = (PDGIDs.TauPrime, PDGIDs.BPrimeQuark, PDGIDs.TPrimeQuark)

    for id in _J_eq_0:
        assert j_spin(id) == 1
    for id in _J_eq_1:
        assert j_spin(id) == 3
    for id in _J_eq_1over2:
        assert j_spin(id) == 2
    for id in _J_eq_3over2:
        assert j_spin(id) == 4
    for id in _invalid_pdgids:
        assert j_spin(id) is None
    for id in _J_eq_None:
        assert j_spin(id) is None


def test_S_non_mesons(PDGIDs):
    _S_eq_None = (PDGIDs.Gluon, PDGIDs.Photon, PDGIDs.Z0)
    for id in _S_eq_None:
        assert S(id) is None


def test_L_non_mesons(PDGIDs):
    _L_eq_None = (PDGIDs.Gluon, PDGIDs.Photon, PDGIDs.Z0)
    for id in _L_eq_None:
        assert L(id) is None


def test_A(PDGIDs):
    _nuclei = {
        PDGIDs.Proton: 1,
        PDGIDs.AntiNeutron: 1,
        PDGIDs.HydrogenNucleus: 1,
        PDGIDs.Carbon12: 12,
    }
    _non_nuclei = [id for id in PDGIDs if id not in _nuclei.keys()]
    for id, a in _nuclei.items():
        assert A(id) == a
    for id in _non_nuclei:
        assert A(id) is None


def test_Z(PDGIDs):
    _nuclei = {
        PDGIDs.Proton: 1,
        PDGIDs.AntiNeutron: 0,
        PDGIDs.HydrogenNucleus: 1,
        PDGIDs.Carbon12: 6,
    }
    _non_nuclei = [id for id in PDGIDs if id not in _nuclei.keys()]
    for id, z in _nuclei.items():
        assert Z(id) == z
    for id in _non_nuclei:
        assert Z(id) is None



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

r"""
The Particle Data Group (PDG) defines the standard particle identification numbering scheme
in the form of a signed 7-digit number +/- N Nr Nl Nq1 Nq2 Nq3 Nj.

PDG IDs with more than 7 digits exist for non-standard particles such as Q-balls.

This module provides the following:

- A pythonic version of the functions defined in HepPID and HepPDT,
  which work with PDG particle identification codes (PDG IDs).
- A handy `PDGID` class.
- A few other functions extending the functionality of the HepXXX code.


Matrix of spin states for mesons
--------------------------------

The table below provides a handy overview of the "links" between PDG IDs
and meson quantum numbers. It is provided here for reference (for lack of a better place!),
as it is not at all easy to find ... not to say impossible.

Useful definitions:

- State: (2S+1)^L_J
- Parity P = (-1)^(L+1)
- Charge conjugation C = (-1)^(L+S), valid only for neutral mesons
- "q q" stands for "nq2 nq3" in the table below

==== === === ======================== ======== ================= ====================== =======================
 J    S   L   State                    J^PC     Name              Mesons                 PDG nL nq1 nq2 nq3 nJ
---- --- --- ------------------------ -------- ----------------- ---------------------- -----------------------
 0    0   0   L=J   , S=0  (1^S_0)     0^-+     pseudo-scalar     pi  eta  eta'  K        00qq1
 0    1   \-  L=J-1 , S=1              \-       \-                \-                      \-
 0    1   \-  L=J   , S=1              \-       \-                \-                      \-
 0    1   1   L=J+1 , S=1  (3^P_0)     0^++     scalar            a_0  f_0  f'_0  K*_0    10qq1
 1    0   1   L=J   , S=0  (1^P_1)     1^+-     pseudo-vector     b_1  h_1  h'_1  K_1     10qq3
 1    1   0   L=J-1 , S=1  (3^S_1)     1^--     vector            rho  omega  phi  K*     00qq3
 1    1   1   L=J   , S=1  (3^P_1)     1^++     axial vector      a_1  f_1  f'_1  K_1     20qq3
 1    1   2   L=J+1 , S=1              1^--     \-                                        30qq3
 2    0   2   L=J   , S=0              2^-+     \-                                        10qq5
 2    1   1   L=J-1 , S=1              2^++     \-                                        00qq5
 2    1   2   L=J   , S=1              2^--     \-                                        20qq5
 2    1   3   L=J+1 , S=1  (3^P_2)     2^++     vector            a_2  f_2  f'_2  K*_2    30qq5
 3    0   3   L=J   , S=0              3^+-     \-                                        10qq7
 3    1   2   L=J-1 , S=1              3^--     \-                                        00qq7
 3    1   3   L=J   , S=1              3^++     \-                                        20qq7
 3    1   4   L=J+1 , S=1              3^--     \-                                        30qq7
 4    0   4   L=J   , S=0              4^-+     \-                                        10qq9
 4    1   3   L=J-1 , S=1              4^++     \-                                        00qq9
 4    1   4   L=J   , S=1              4^--     \-                                        20qq9
 4    1   5   L=J+1 , S=1              4^++     \-                                        30qq9
==== === === ======================== ======== ================= ====================== =======================

"""


from __future__ import annotations

from .functions import (
    A,
    J,
    L,
    S,
    Z,
    abspid,
    charge,
    has_bottom,
    has_charm,
    has_down,
    has_fundamental_anti,
    has_strange,
    has_top,
    has_up,
    is_baryon,
    is_diquark,
    is_dyon,
    is_excited_quark_or_lepton,
    is_gauge_boson_or_higgs,
    is_generator_specific,
    is_hadron,
    is_lepton,
    is_meson,
    is_nucleus,
    is_pentaquark,
    is_Qball,
    is_quark,
    is_Rhadron,
    is_sm_gauge_boson_or_higgs,
    is_sm_lepton,
    is_sm_quark,
    is_special_particle,
    is_SUSY,
    is_technicolor,
    is_valid,
    j_spin,
    l_spin,
    s_spin,
    three_charge,
)
from .pdgid import PDGID

__all__ = (
    "PDGID",
    #
    "is_valid",
    "abspid",
    #
    "is_Qball",
    "is_Rhadron",
    "is_SUSY",
    "is_baryon",
    "is_diquark",
    "is_dyon",
    "is_excited_quark_or_lepton",
    "is_gauge_boson_or_higgs",
    "is_generator_specific",
    "is_hadron",
    "is_lepton",
    "is_meson",
    "is_nucleus",
    "is_pentaquark",
    "is_quark",
    "is_sm_gauge_boson_or_higgs",
    "is_sm_lepton",
    "is_sm_quark",
    "is_special_particle",
    "is_technicolor",
    #
    "has_down",
    "has_up",
    "has_strange",
    "has_charm",
    "has_bottom",
    "has_top",
    "has_fundamental_anti",
    #
    "charge",
    "three_charge",
    "j_spin",
    "J",
    "s_spin",
    "S",
    "l_spin",
    "L",
    "A",
    "Z",
)


def __dir__() -> tuple[str, ...]:
    return __all__



#!/usr/bin/env python
# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.
"""This script generates the PDGID<->LHCb name mapping using the table from the LHCb DDDB package."""

from __future__ import annotations

import csv
import datetime as dt

import requests


def download_table(
    url="https://gitlab.cern.ch/lhcb-conddb/DDDB/-/raw/master/param/ParticleTable.txt",
):
    r = requests.get(url)
    r.raise_for_status()

    lines = r.text.split("\n")

    return [line.split() for line in filter(lambda x: x and x[0] == " ", lines)]


def main():
    date = dt.datetime.today().strftime("%Y-%m-%d")
    table = download_table()
    lhcb_names = {int(pdg_id): name for name, _, pdg_id, *_ in table}

    with open("src/particle/lhcb/data/pdgid_to_lhcbname.csv", "w") as out_csv:
        out_csv.write(
            f"# (c) Scikit-HEP project - Particle package data file - pdgid_to_lhcbname.csv - {date}\n"
        )
        writer = csv.DictWriter(
            out_csv, fieldnames=("PDGID", "STR"), lineterminator="\n"
        )
        writer.writeheader()

        for pid, name in sorted(
            lhcb_names.items(), key=lambda x: (abs(int(x[0])), -int(x[0]))
        ):
            writer.writerow({"PDGID": pid, "STR": name})


if __name__ == "__main__":
    main()



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

from ..particle import Particle
from .converters import LHCbName2PDGIDBiMap


def to_lhcb_name(p: Particle) -> str:
    """
    Convert to the name used in the LHCb software framework.

    Examples
    --------
    >>> p = Particle.from_pdgid(-531)
    >>> p
    <Particle: name="B(s)~0", pdgid=-531, mass=5366.88 ± 0.14 MeV>
    >>> to_lhcb_name(p)
    'B_s~0'
    """
    return LHCbName2PDGIDBiMap[p.pdgid]


def from_lhcb_name(name: str) -> Particle:
    """
    Get a `Particle` from an LHCb particle name, as used in LHCb Gaudi applications.

    Examples
    --------
    >>> from_lhcb_name("B_s~0")
    <Particle: name="B(s)~0", pdgid=-531, mass=5366.88 ± 0.14 MeV>

    Raises
    ------
    ParticleNotFound
        If `from_pdgid` returns no match.
    MatchingIDNotFound
        If the matching LHCb name - PDG ID done internally is unsuccessful.
    """
    return Particle.from_pdgid(LHCbName2PDGIDBiMap[name])



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
The Particle Data Group (PDG) defines the standard particle identification numbering scheme
in the form of a signed 7-digit number +/- N Nr Nl Nq1 Nq2 Nq3 Nj.

PDG IDs with more than 7 digits exist for non-standard particles such as Q-balls.
These follow outside the standard PDG numbering scheme.

This module provides the following:

- A pythonic version of the functions defined in HepPID and HepPDT,
  which work with PDG particle identification codes (PDG IDs).
- A few other functions extending the functionality of the HepXXX code.

References
----------

- PDG document "Monte Carlo Particle Numbering Scheme".
- HepPDT and HepPID versions 3.04.01.
"""


from __future__ import annotations

from enum import IntEnum
from typing import SupportsInt

PDGID_TYPE = SupportsInt


class Location(IntEnum):
    """
    Represents the location digit in the PDG numbering scheme, which provides a convenient index into the PID.
    """

    Nj = 1
    Nq3 = 2
    Nq2 = 3
    Nq1 = 4
    Nl = 5
    Nr = 6
    N = 7
    N8 = 8
    N9 = 9
    N10 = 10


def is_valid(pdgid: PDGID_TYPE) -> bool:
    """Is it a valid PDG ID?"""
    if _fundamental_id(pdgid) != 0:  # function always returns a number >= 0
        return True
    if is_meson(pdgid):
        return True
    if is_baryon(pdgid):
        return True
    if is_gauge_boson_or_higgs(pdgid):
        return True
    if is_pentaquark(pdgid):
        return True
    if is_SUSY(pdgid):
        return True
    if is_Rhadron(pdgid):
        return True
    if is_dyon(pdgid):
        return True
    if is_diquark(pdgid):
        return True
    if is_pentaquark(pdgid):
        return True
    if is_generator_specific(pdgid):
        return True
    if is_technicolor(pdgid):
        return True
    if is_excited_quark_or_lepton(pdgid):
        return True
    if _extra_bits(pdgid) > 0:
        return is_Qball(pdgid) or is_nucleus(pdgid)
    return False


def abspid(pdgid: PDGID_TYPE) -> int:
    """Returns the absolute value of the PDG ID."""
    return abs(int(pdgid))


def is_quark(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a quark?

    Standard Model and fourth-generation quarks are included, but not excited (composite) quarks.

    Codes 1-8 are reserved for Standard Model or fourth-generation quarks,
    and exclude codes for excited, i.e. composite, quarks (see ``is_excited_quark_or_lepton``).
    """
    return 1 <= abspid(pdgid) <= 8


def is_sm_quark(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a Standard Model quark?

    Fourth-generation quarks are not included.

    Codes 1-8 are reserved for Standard Model (SM) or fourth-generation quarks,
    and exclude codes for excited, i.e. composite, quarks (see ``is_excited_quark_or_lepton``),
    but only the codes 1-6 actually correspond to SM quarks.
    """
    return 1 <= abspid(pdgid) <= 6


def is_lepton(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a lepton?

    Standard Model and fourth-generation leptons are included, but not excited (composite) leptons.

    Codes 11-18 are reserved for Standard Model or fourth-generation leptons,
    and exclude codes for excited, i.e. composite, leptons (see ``is_excited_quark_or_lepton``).
    """
    return 11 <= abspid(pdgid) <= 18


def is_sm_lepton(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a Standard Model lepton?

    Fourth-generation leptons are not included.

    Codes 11-18 are reserved for Standard Model (SM) or fourth-generation leptons,
    and exclude codes for excited, i.e. composite, leptons (see ``is_excited_quark_or_lepton``),
    but only the codes 11-16 actually correspond to SM leptons.
    """
    return 11 <= abspid(pdgid) <= 16


def is_hadron(pdgid: PDGID_TYPE) -> bool:
    """Does this PDG ID correspond to a hadron?"""
    # Special case of proton and neutron:
    # needs to be checked first since _extra_bits(pdgid) > 0 for nuclei
    if abs(int(pdgid)) in {1000000010, 1000010010}:
        return True
    if _extra_bits(pdgid) > 0:
        return False
    if is_meson(pdgid):
        return True
    if is_baryon(pdgid):
        return True
    if is_pentaquark(pdgid):
        return True
    return bool(is_Rhadron(pdgid))


def is_meson(pdgid: PDGID_TYPE) -> bool:
    """Does this PDG ID correspond to a meson?"""
    if _extra_bits(pdgid) > 0:
        return False
    if abspid(pdgid) <= 100:
        return False
    if 0 < int(_fundamental_id(pdgid)) <= 100:
        return False
    # Special IDs - K(L)0, ???, K(S)0
    if abspid(pdgid) in {130, 210, 310}:
        return True
    # Special IDs - B(L)0, B(sL)0, B(H)0, B(sH)0
    if abspid(pdgid) in {150, 350, 510, 530}:
        return True
    # Special particles - reggeon, pomeron, odderon
    if int(pdgid) in {110, 990, 9990}:
        return True
    if (
        _digit(pdgid, Location.Nj) > 0
        and _digit(pdgid, Location.Nq3) > 0
        and _digit(pdgid, Location.Nq2) > 0
        and _digit(pdgid, Location.Nq1) == 0
    ):
        # check for illegal antiparticles
        return (
            _digit(pdgid, Location.Nq3) != _digit(pdgid, Location.Nq2)
            or int(pdgid) >= 0
        )

    return False


def is_baryon(pdgid: PDGID_TYPE) -> bool:
    """Does this PDG ID correspond to a baryon?"""
    if abspid(pdgid) <= 100:
        return False
    # Special case of proton and neutron:
    # needs to be checked first since _extra_bits(pdgid) > 0 for nuclei
    if abs(int(pdgid)) in {1000000010, 1000010010}:
        return True

    if _extra_bits(pdgid) > 0:
        return False

    if 0 < _fundamental_id(pdgid) <= 100:
        return False

    # Old codes for diffractive p and n (MC usage)
    if abspid(pdgid) in {2110, 2210}:
        return True

    return (
        _digit(pdgid, Location.Nj) > 0
        and _digit(pdgid, Location.Nq3) > 0
        and _digit(pdgid, Location.Nq2) > 0
        and _digit(pdgid, Location.Nq1) > 0
    )


def is_diquark(pdgid: PDGID_TYPE) -> bool:
    """Does this PDG ID correspond to a diquark?"""
    if _extra_bits(pdgid) > 0:
        return False
    if abspid(pdgid) <= 100:
        return False
    if 0 < int(_fundamental_id(pdgid)) <= 100:
        return False
    return (
        _digit(pdgid, Location.Nj) > 0
        and _digit(pdgid, Location.Nq3) == 0
        and _digit(pdgid, Location.Nq2) > 0
        and _digit(pdgid, Location.Nq1) > 0
    )


def is_nucleus(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a nucleus?

    Ion numbers are +/- 10LZZZAAAI.
    AAA is A - total baryon number
    ZZZ is Z - total charge
    L is the total number of strange quarks.
    I is the isomer number, with I=0 corresponding to the ground state.
    """
    # A proton can be a Hydrogen nucleus
    # A neutron can be considered as a nucleus when given the PDG ID 1000000010,
    # hence consistency demands that is_nucleus(neutron) is True
    if abspid(pdgid) in {2112, 2212}:
        return True
    if _digit(pdgid, Location.N10) == 1 and _digit(pdgid, Location.N9) == 0:
        # Charge should always be less than or equal to the baryon number
        A_pdgid = A(pdgid)
        Z_pdgid = Z(pdgid)

        if A_pdgid is None or Z_pdgid is None:
            return False
        if A_pdgid >= abs(Z_pdgid):
            return True
    return False


def is_pentaquark(pdgid: PDGID_TYPE) -> bool:
    """
    Does the PDG ID correspond to a pentaquark?

    Pentaquark IDs are of the form +/- 9 Nr Nl Nq1 Nq2 Nq3 Nj, where Nj = 2J + 1 gives the spin
    and Nr Nl Nq1 Nq2 Nq3 denote the quark numbers in order Nr >= Nl >= Nq1 >= Nq2
    and Nq3 gives the antiquark number.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) != 9:
        return False
    if _digit(pdgid, Location.Nr) in {9, 0}:
        return False
    if _digit(pdgid, Location.Nj) == 9 or _digit(pdgid, Location.Nl) == 0:
        return False
    if _digit(pdgid, Location.Nq1) == 0:
        return False
    if _digit(pdgid, Location.Nq2) == 0:
        return False
    if _digit(pdgid, Location.Nq3) == 0:
        return False
    if _digit(pdgid, Location.Nj) == 0:
        return False
    if _digit(pdgid, Location.Nq2) > _digit(pdgid, Location.Nq1):
        return False
    if _digit(pdgid, Location.Nq1) > _digit(pdgid, Location.Nl):
        return False
    return _digit(pdgid, Location.Nl) <= _digit(pdgid, Location.Nr)


def is_gauge_boson_or_higgs(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a gauge boson or a Higgs?

    Codes 21-30 are reserved for the Standard Model gauge bosons and the Higgs.
    The graviton and the boson content of a two-Higgs-doublet scenario
    and of additional SU(2)xU(1) groups are found in the range 31-40.
    """
    return 21 <= abspid(pdgid) <= 40


def is_sm_gauge_boson_or_higgs(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a Standard Model gauge boson or Higgs?

    Codes 21-30 are reserved for the Standard Model gauge bosons and the Higgs,
    but only the codes 21-25 actually correspond to SM particles.
    """
    if abspid(pdgid) == 24:  # W is the only SM gauge boson not its antiparticle
        return True

    return 21 <= int(pdgid) <= 25


def is_generator_specific(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to generator-specific pseudoparticles or concepts?

    Codes 81-100 are reserved for generator-specific pseudoparticles and concepts.
    Codes 901-930, 1901-1930, 2901-2930, and 3901-3930 are for
    additional components of Standard Model parton distribution functions,
    where the latter three ranges are intended to distinguish
    left/right/longitudinal components.
    Codes 998 and 999 are reserved for GEANT tracking purposes.
    """
    aid = abspid(pdgid)
    if 81 <= aid <= 100:
        return True
    if 901 <= aid <= 930:
        return True
    if 1901 <= aid <= 1930:
        return True
    if 2901 <= aid <= 2930:
        return True
    if 3901 <= aid <= 3930:
        return True
    if aid in {998, 999}:
        return True
    if aid in {20022, 480000000}:  # Special cases of opticalphoton and geantino
        return True
    return False


def is_special_particle(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a special particle?

    Special particle in the sense of the classification in the PDG MC particle numbering scheme document,
    hence the graviton, the DM (S = 0, 1/2, 1) particles, the reggeons (reggeon, pomeron and odderon),
    and all generator-specific pseudo-particles and concepts, see `is_generator_specific`.
    """
    return pdgid in {39, 41, 42, 51, 52, 53, 110, 990, 9990} or is_generator_specific(
        pdgid
    )


def is_Rhadron(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to an R-hadron?

    An R-hadron is of the form 10abcdj, 100abcj, or 1000abj,
    where j = 2J + 1 gives the spin; b, c, and d are quarks or gluons;
    and a (the digit following the zero's) is a SUSY particle.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) != 1:
        return False
    if _digit(pdgid, Location.Nr) != 0:
        return False
    if is_SUSY(pdgid):
        return False
    # All R-hadrons have at least 3 core digits
    return (
        _digit(pdgid, Location.Nq2) != 0
        and _digit(pdgid, Location.Nq3) != 0
        and _digit(pdgid, Location.Nj) != 0
    )


def is_Qball(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a Q-ball or any exotic particle with electric charge beyond the qqq scheme?

    Ad-hoc numbering for such particles is +/- 100XXXY0, where XXX.Y is the charge.
    """
    if _extra_bits(pdgid) != 1:
        return False
    if _digit(pdgid, Location.N) != 0:
        return False
    if _digit(pdgid, Location.Nr) != 0:
        return False
    if (abspid(pdgid) // 10) % 10000 == 0:
        return False
    return _digit(pdgid, Location.Nj) == 0


def is_dyon(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a Dyon, a magnetic monopole?

    Magnetic monopoles and Dyons are assumed to have one unit of Dirac monopole charge
    and a variable integer number xyz units of electric charge,
    where xyz stands for Nq1 Nq2 Nq3.

    Codes 411xyz0 are used when the magnetic and electrical charge sign agree and 412xyz0 when they disagree,
    with the overall sign of the particle set by the magnetic charge.
    For now, no spin information is provided.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) != 4:
        return False
    if _digit(pdgid, Location.Nr) != 1:
        return False
    if _digit(pdgid, Location.Nl) not in {1, 2}:
        return False
    if _digit(pdgid, Location.Nq3) == 0:
        return False
    return _digit(pdgid, Location.Nj) == 0


def is_SUSY(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a SUSY particle?

    Fundamental SUSY particles have N = 1 or 2.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _digit(pdgid, Location.N) not in {1, 2}:
        return False
    if _digit(pdgid, Location.Nr) != 0:
        return False
    return _fundamental_id(pdgid) != 0


def is_technicolor(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to a Technicolor state?

    Technicolor states have N = 3.
    """
    if _extra_bits(pdgid) > 0:
        return False
    return _digit(pdgid, Location.N) == 3


def is_excited_quark_or_lepton(pdgid: PDGID_TYPE) -> bool:
    """
    Does this PDG ID correspond to an excited (composite) quark or lepton?

    Excited (composite) quarks and leptons have N = 4 and Nr = 0.
    """
    if _extra_bits(pdgid) > 0:
        return False
    if _fundamental_id(pdgid) == 0:
        return False
    return _digit(pdgid, Location.N) == 4 and _digit(pdgid, Location.Nr) == 0


def has_down(pdgid: PDGID_TYPE) -> bool:
    """Does this particle contain a down quark?"""
    return _has_quark_q(pdgid, 1)


def has_up(pdgid: PDGID_TYPE) -> bool:
    """Does this particle contain an up quark?"""
    return _has_quark_q(pdgid, 2)


def has_strange(pdgid: PDGID_TYPE) -> bool:
    """Does this particle contain a strange quark?"""
    return _has_quark_q(pdgid, 3)


def has_charm(pdgid: PDGID_TYPE) -> bool:
    """Does this particle contain a charm quark?"""
    return _has_quark_q(pdgid, 4)


def has_bottom(pdgid: PDGID_TYPE) -> bool:
    """Does this particle contain a bottom quark?"""
    return _has_quark_q(pdgid, 5)


def has_top(pdgid: PDGID_TYPE) -> bool:
    """Does this particle contain a top quark?"""
    return _has_quark_q(pdgid, 6)


def has_fundamental_anti(pdgid: PDGID_TYPE) -> bool:
    """
    If this is a fundamental particle, does it have a valid antiparticle?

    Notes
    -----
    Based on the current list of defined particles/concepts
    in the PDG Monte Carlo Particle Numbering Scheme document.
    """
    fid = _fundamental_id(pdgid)  # always a positive integer

    # Check generator-specific PDGIDs
    if 81 <= fid <= 100:
        return fid in {82, 84, 85, 86, 87}

    # Check PDGIDs from 1 to 79
    _cp_conjugates = {21, 22, 23, 25, 32, 33, 35, 36, 39, 40, 43}
    _unassigned = (
        [9, 10, 19, 20, 26] + list(range(26, 32)) + list(range(45, 80))
    )  # not in conversion.csv
    if (1 <= fid <= 79) and fid not in _cp_conjugates:
        return fid not in _unassigned

    return False


def charge(pdgid: PDGID_TYPE) -> float | None:
    """Returns the charge."""

    three_charge_pdgid = three_charge(pdgid)
    if three_charge_pdgid is None:
        return None
    if not is_Qball(pdgid):
        return three_charge_pdgid / 3.0
    return three_charge_pdgid / 30.0


def three_charge(pdgid: PDGID_TYPE) -> int | None:
    """
    Returns 3 times the charge.

    None is returned if the PDGID is not valid.
    """
    if not is_valid(pdgid):
        return None

    aid = abspid(pdgid)
    charge = None  # pylint: disable=redefined-outer-name
    ch100 = [
        -1,
        2,
        -1,
        2,
        -1,
        2,
        -1,
        2,
        0,
        0,
        -3,
        0,
        -3,
        0,
        -3,
        0,
        -3,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        0,
        0,
        3,
        0,
        0,
        0,
        0,
        -1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        6,
        3,
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]
    q1 = _digit(pdgid, Location.Nq1)
    q2 = _digit(pdgid, Location.Nq2)
    q3 = _digit(pdgid, Location.Nq3)
    sid = _fundamental_id(pdgid)

    if _extra_bits(pdgid) > 0:
        if is_nucleus(pdgid):  # ion
            Z_pdgid = Z(pdgid)
            if Z_pdgid is None:
                return None
            return 3 * Z_pdgid
        if is_Qball(pdgid):  # Qball
            charge = 3 * ((aid // 10) % 10000)
        else:  # this should never be reached in the present numbering scheme
            return None  # since extra bits exist only for Q-balls and nuclei
    elif is_dyon(pdgid):  # Dyon
        charge = 3 * ((aid // 10) % 1000)
        # this is half right
        # the charge sign will be changed below if pid < 0
        if _digit(pdgid, Location.Nl) == 2:
            charge = -charge
    elif 0 < sid <= 100:  # use table
        charge = ch100[sid - 1]
        if aid in {1000017, 1000018, 1000034, 1000052, 1000053, 1000054}:
            charge = 0
        if aid in {5100061, 5100062}:
            charge = 6
    elif _digit(pdgid, Location.Nj) == 0:  # KL, KS, or undefined
        return 0
    elif q1 == 0 or (is_Rhadron(pdgid) and q1 == 9):  # mesons
        if q2 in {3, 5}:
            charge = ch100[q3 - 1] - ch100[q2 - 1]
        else:
            charge = ch100[q2 - 1] - ch100[q3 - 1]
    elif q3 == 0:  # diquarks
        charge = ch100[q2 - 1] + ch100[q1 - 1]
    elif is_baryon(pdgid) or (
        is_Rhadron(pdgid) and _digit(pdgid, Location.Nl) == 9
    ):  # baryons
        charge = ch100[q3 - 1] + ch100[q2 - 1] + ch100[q1 - 1]

    if charge is not None and int(pdgid) < 0:
        charge = -charge
    return charge


def j_spin(pdgid: PDGID_TYPE) -> int | None:
    """Returns the total spin as 2J+1."""
    if not is_valid(pdgid):
        return None
    if _fundamental_id(pdgid) > 0:
        fund = _fundamental_id(pdgid)
        if 0 < fund < 7:
            return 2  # 4th generation quarks not dealt with !
        if fund == 9:
            return 3
        if 10 < fund < 17:
            return 2  # 4th generation leptons not dealt with !
        if 20 < fund < 25:
            return 3
        return None
    if abs(int(pdgid)) in {1000000010, 1000010010}:  # neutron, proton
        return 2
    if _extra_bits(pdgid) > 0:
        return None
    if pdgid in {130, 310}:
        return 1  # Special cases of the KS and KL !
    return abspid(pdgid) % 10


def J(pdgid: PDGID_TYPE) -> float | None:
    """Returns the total spin J."""
    value = j_spin(pdgid)
    return (value - 1) / 2 if value is not None else value


def S(pdgid: PDGID_TYPE) -> int | None:
    """
    Returns the spin S.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    if not is_meson(pdgid):
        return None

    if not is_valid(pdgid):
        return None

    if (abspid(pdgid) // 1000000) % 10 == 9:
        return None  # no knowledge so far

    nl = (abspid(pdgid) // 10000) % 10
    js = abspid(pdgid) % 10

    if js != 1 and js < 3:
        return 0

    if nl == 0:
        return 0 if js == 1 else 1
    if nl == 1:
        return 1 if js == 1 else 0
    if nl in {2, 3}:
        return 1 if js >= 3 else 0

    return 0


def s_spin(pdgid: PDGID_TYPE) -> int | None:
    """
    Returns the spin S as 2S+1.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    value = S(pdgid)
    return (2 * value + 1) if value is not None else value


def L(pdgid: PDGID_TYPE) -> int | None:
    """
    Returns the orbital angular momentum L.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    if not is_meson(pdgid):
        return None

    if not is_valid(pdgid):
        return None

    if (abspid(pdgid) // 1000000) % 10 == 9:
        return None  # no knowledge so far

    nl = (abspid(pdgid) // 10000) % 10
    js = abspid(pdgid) % 10

    if nl == 0:
        if js in {1, 3}:
            return 0
        if js == 5:
            return 1
        if js == 7:
            return 2
        if js == 9:
            return 3
    elif nl == 1:
        if js in {1, 3}:
            return 1
        if js == 5:
            return 2
        if js == 7:
            return 3
        if js == 9:
            return 4
    elif nl == 2:
        if js == 3:
            return 1
        if js == 5:
            return 2
        if js == 7:
            return 3
        if js == 9:
            return 4
    elif nl == 3:
        if js == 3:
            return 2
        if js == 5:
            return 3
        if js == 7:
            return 4
        if js == 9:
            return 5

    return 0


def l_spin(pdgid: PDGID_TYPE) -> int | None:
    """
    Returns the orbital angular momentum L as 2L+1.

    Notes
    -----
    - This is valid for mesons only. None is returned otherwise.
    - Mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
      and None is returned too.
    """
    value = L(pdgid)
    return (2 * value + 1) if value is not None else value


def A(pdgid: PDGID_TYPE) -> int | None:
    """
    Returns the atomic mass number A if the PDG ID corresponds to a nucleus, else it returns None.

    The (atomic) mass number is also known as the nucleon number, the total number of protons and neutrons.
    The number of neutrons is hence N = A - Z.
    """
    # A proton can be a Hydrogen nucleus
    # A neutron can be considered as a nucleus when given the PDG ID 1000000010,
    # hence consistency demands that A(neutron) = 1
    if abspid(pdgid) in {2112, 2212}:
        return 1
    if _digit(pdgid, Location.N10) != 1 or _digit(pdgid, Location.N9) != 0:
        return None
    return (abspid(pdgid) // 10) % 1000


def Z(pdgid: PDGID_TYPE) -> int | None:
    """
    Returns the nuclear charge Z if the PDG ID corresponds to a nucleus, else it returns None.

    The nuclear charge number Z is also known as the atomic number.
    For ordinary nuclei Z is equal to the number of protons.
    """
    # A proton can be a Hydrogen nucleus
    if abspid(pdgid) == 2212:
        return int(pdgid) // 2212
    # A neutron can be considered as a nucleus when given the PDG ID 1000000010,
    # hence consistency demands that Z(neutron) = 0
    if abspid(pdgid) == 2112:
        return 0
    if _digit(pdgid, Location.N10) != 1 or _digit(pdgid, Location.N9) != 0:
        return None
    return ((abspid(pdgid) // 10000) % 1000) * (int(pdgid) // abs(int(pdgid)))


def _digit(pdgid: PDGID_TYPE, loc: int) -> int:
    """
    Provides a convenient index into the PDGID number, whose format is in base 10.

    Returns the digit at position 'loc' given that the right-most digit is at position 1.
    """
    sid = str(abspid(pdgid))
    return int(sid[-loc]) if loc <= len(sid) else 0


def _extra_bits(pdgid: PDGID_TYPE) -> int:
    """
    Returns everything beyond the 7th digit, so anything outside the PDG numbering scheme.
    """
    return abspid(pdgid) // 10000000


def _fundamental_id(pdgid: PDGID_TYPE) -> int:
    """
    Returns the first 2 digits if this is a "fundamental" particle.
    Returns 0 if the particle is not fundamental or not standard (PDG ID with more than 7 digits).

    PDGID=100 is a special case (internal generator ID's are 81-100).

    Notes
    -----
    Function always returns a number >= 0.
    """
    if _extra_bits(pdgid) > 0:
        return 0
    if abspid(pdgid) <= 100:
        return abspid(pdgid)
    if _digit(pdgid, Location.Nq2) == _digit(pdgid, Location.Nq1) == 0:
        return abspid(pdgid) % 10000
    return 0


def _has_quark_q(pdgid: PDGID_TYPE, q: int) -> bool:
    """
    Helper function - does this particle contain a quark q?

    Note that q is always positive, so [1, 6] for Standard Model quarks
    and [7, 8] for fourth-generation quarks.
    """
    # Nuclei can also contain strange quarks,
    # cf. the definition of a nucleus PDG ID in is_nucleus.
    # This check needs to be done first since _extra_bits(pdgid) > 0 for nuclei
    if is_nucleus(pdgid):
        if q in {1, 2}:
            return True  # Nuclei by construction contain up and down quarks
        if q == 3 and pdgid not in {2112, 2212}:
            return _digit(pdgid, Location.N8) > 0
    if _extra_bits(pdgid) > 0:
        return False
    if _fundamental_id(pdgid) > 0:
        return False

    if is_dyon(pdgid):
        return False

    if is_Rhadron(pdgid):
        iz = 7
        for loc in range(6, 1, -1):
            if _digit(pdgid, loc) == 0:
                iz = loc
            elif loc != iz - 1:
                if _digit(pdgid, loc) == q:
                    return True
            # ignore squark or gluino
        return False

    if (
        _digit(pdgid, Location.Nq3) == q
        or _digit(pdgid, Location.Nq2) == q
        or _digit(pdgid, Location.Nq1) == q
    ):
        return True

    if is_pentaquark(pdgid) and (
        _digit(pdgid, Location.Nl) == q or _digit(pdgid, Location.Nr) == q
    ):
        return True

    return False



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Class representing a Pythia ID.
"""


from __future__ import annotations

import csv
from typing import TypeVar

from .. import data
from ..exceptions import MatchingIDNotFound
from ..pdgid import PDGID

with data.basepath.joinpath("pdgid_to_pythiaid.csv").open() as _f:
    _bimap = {
        int(v["PYTHIAID"]): int(v["PDGID"])
        for v in csv.DictReader(line for line in _f if not line.startswith("#"))
    }


Self = TypeVar("Self", bound="PythiaID")


class PythiaID(int):
    """
    Holds a Pythia ID.

    Examples
    --------
    >>> pythiaid = PythiaID(211)

    >>> from particle import Particle
    >>> p = Particle.from_pdgid(pythiaid.to_pdgid())

    >>> (p,) = Particle.finditer(pdgid=pythiaid.to_pdgid())
    >>> p.name
    'pi+'
    """

    __slots__ = ()  # Keep PythiaID a slots based class

    @classmethod
    def from_pdgid(cls: type[Self], pdgid: int) -> Self:
        """
        Constructor from a PDGID.
        """
        for k, v in _bimap.items():
            if v == pdgid:
                return cls(k)
        raise MatchingIDNotFound(f"Non-existent PythiaID for input PDGID {pdgid} !")

    def to_pdgid(self) -> PDGID:
        return PDGID(_bimap[self])

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {int(self):d}>"

    def __str__(self) -> str:
        return repr(self)

    def __neg__(self: Self) -> Self:
        return self.__class__(-int(self))

    __invert__ = __neg__



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

from ..converters.bimap import BiMap
from ..pdgid import PDGID
from . import data

LHCbName2PDGIDBiMap = BiMap(
    PDGID,
    str,
    converters=(int, str),
    filename=data.basepath / "pdgid_to_lhcbname.csv",
)
LHCbName2PDGIDBiMap.__doc__ = """
Bi-bidirectional map between PDG IDs and LHCb names.

Examples
--------
>>> name = LHCbName2PDGIDBiMap[PDGID(-531)]
>>> name
'B_s~0'

>>> pdgid = LHCbName2PDGIDBiMap['B_s~0']
>>> pdgid
<PDGID: -531>
"""



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Functions relevant to particle kinematics.
"""


from __future__ import annotations

from hepunits.constants import hbar
from hepunits.units import MeV, ns


def width_to_lifetime(Gamma: float) -> float:
    """
    Convert from a particle decay width to a lifetime.

    Parameters
    ----------
    Gamma : float > 0
        Particle decay width, in the HEP standard energy unit MeV.

    Returns
    -------
    Gamma > 0: particle lifetime, in the HEP standard time unit ns.
    Gamma = 0: Infinity (float("inf")).
    Gamma < 0: an exception ValueError is raised.

    Examples
    --------
    Manipulation with no explicit usage of units:

    >>> width_to_lifetime(4.33e-10)   # result returned in ns
    0.0015201199929582136

    Manipulations with explicit units defined in the HEP system of units:

    >>> from hepunits.units import MeV, eV, ps   # handy module with units in the HEP system of units
    >>>
    >>> width_to_lifetime(4.33e-10*MeV)   # result returned in ns
    0.0015201199929582136
    >>>
    >>> width_to_lifetime(4.33e-4*eV)   # result again returned in ns
    0.0015201199929582136
    >>>
    >>> width_to_lifetime(4.33e-10*MeV)/ps   # result converted to ps
    1.5201199929582137
    """

    if Gamma < 0.0:
        raise ValueError(f"Input provided, {Gamma} <= 0!")
    if Gamma == 0:
        return float("inf")

    # Just need to first make sure that the width is in the standard unit MeV
    return hbar / float(Gamma / MeV)


def lifetime_to_width(tau: float) -> float:
    """
    Convert from a particle lifetime to a decay width.

    Parameters
    -----------
    tau : float > 0
        Particle lifetime, in the HEP standard time unit ns.

    Returns
    -------
    Particle decay width, in the HEP standard energy unit MeV.
    tau > 0: particle lifetime, in the HEP standard time unit ns.
    tau = 0: Infinity (float("inf")).
    tau < 0: an exception ValueError is raised.

    Examples
    --------
    Manipulation with no explicit usage of units:

    >>> lifetime_to_width(0.0015201199929582136)   # result returned in MeV
    4.33e-10

    Manipulations with explicit units defined in the HEP system of units:

    >>> from hepunits.units import MeV, eV, ps   # handy module with units in the HEP system of units
    >>>
    >>> lifetime_to_width(0.0015201199929582136*ns)   # result returned in MeV
    4.33e-10
    >>>
    >>> lifetime_to_width(1.5201199929582137*ps)   # result again returned in MeV
    4.33e-10
    >>>
    >>> lifetime_to_width(1.5201199929582137*ps)/eV   # result converted to eV
    0.000433
    """

    if tau < 0:
        raise ValueError(f"Input provided, {tau} <= 0!")
    if tau == 0:
        return float("inf")

    # Just need to first make sure that the lifetime is in the standard unit ns
    return hbar / float(tau / ns)



#!/usr/bin/env python
# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

import argparse
import sys

from . import __version__
from .particle import Particle
from .pdgid import PDGID


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="particle",
        description="Particle command line display utility. Has two modes.",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    subparsers = parser.add_subparsers(help="Subcommands")

    search = subparsers.add_parser(
        "search",
        help="Look up particles by PID or name (Ex.: python -m particle search D+ D-)",
    )
    search.add_argument("particle", nargs="+", help="Name(s) or ID(s)")

    pdgid = subparsers.add_parser(
        "pdgid", help="Print info from PID (Ex.: python -m particle pdgid 11 13)"
    )
    pdgid.add_argument("pdgid", nargs="+", help="ID(s)")

    opts = parser.parse_args()

    if "particle" in opts:
        for cand in opts.particle:
            if hasattr(cand, "decode"):
                cand = cand.decode("utf-8")

            try:
                value = int(cand)
            except ValueError:
                value = 0

            if value:
                particles = [Particle.from_pdgid(value)]
            else:
                particles = Particle.from_string_list(cand)

            if not particles:
                print("Particle", cand, "not found.")
                sys.exit(1)
            elif len(particles) == 1:
                print(particles[0].describe())
            else:
                for particle in particles:
                    print(repr(particle))

            print()

    if "pdgid" in opts:
        for value in opts.pdgid:
            p = PDGID(value)
            print(p)
            print(PDGID(value).info())


if __name__ == "__main__":
    main()



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

from particle.exceptions import MatchingIDNotFound
from particle.geant import Geant3ID
from particle.pdgid import PDGID


def test_class_string_representations():
    pid = Geant3ID(1)
    assert pid == 1
    assert pid.__str__() == "<Geant3ID: 1>"


def test_class_return_type():
    assert isinstance(-Geant3ID(3), Geant3ID)
    assert isinstance(~Geant3ID(3), Geant3ID)


def test_class_inversion():
    assert -Geant3ID(1) == ~Geant3ID(1)


def test_from_pdgid():
    assert Geant3ID.from_pdgid(211) == 8

    assert Geant3ID.from_pdgid(PDGID(211)) == 8
    assert Geant3ID.from_pdgid(PDGID(211)) == Geant3ID(8)


def test_from_pdgid_non_matching():
    with pytest.raises(MatchingIDNotFound):
        Geant3ID.from_pdgid(55)


def test_to_pdgid():
    gid = Geant3ID(8)
    assert gid.to_pdgid() == 211
    assert gid.to_pdgid() == PDGID(211)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from .corsika7id import Corsika7ID

__all__ = ("Corsika7ID",)


def __dir__() -> tuple[str, ...]:
    return __all__



from __future__ import annotations

from particle import Particle, data


def test_load_particle_table(benchmark):
    benchmark(Particle.load_table, data.basepath / "particle2022.csv")


def test_load_nuclei_append(benchmark):
    def load_two():
        Particle.load_table(data.basepath / "particle2022.csv")
        Particle.load_table(data.basepath / "nuclei2020.csv", append=True)

    benchmark(load_two)


def test_from_pdgid(benchmark):
    Particle.load_table(data.basepath / "particle2022.csv")
    table = [int(s.pdgid) for s in Particle.all()]

    def get_all(listing):
        for pdgid in listing:
            Particle.from_pdgid(pdgid)

    benchmark(get_all, table)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Collection of enums to help characterising particle properties
Examples are charge, spin and parity.
"""

from __future__ import annotations

from enum import IntEnum


class SpinType(IntEnum):
    """
    Enum representing the spin type. Relevant only for bosons.

    SpinType.Unknown is returned for bosons if one of the values (J,P) is not known/relevant.
    SpinType.NonDefined is to be used for non-bosons.
    """

    #          Values of (J, P)
    Scalar = 1  # (0, 1)
    PseudoScalar = -1  # (0,-1)
    Vector = 2  # (1,-1)
    Axial = -2  # (1, 1)
    Tensor = 3  # (2, 1)
    PseudoTensor = -3  # (2,-1)
    Unknown = 0
    NonDefined = 5


class Parity(IntEnum):
    """Enum representing a particle parity."""

    p = 1
    m = -1
    u = 5


class Charge(IntEnum):
    """Enum representing the particle charge * 3."""

    pp = 6
    p43 = 4  # 4/3
    p = 3
    p23 = 2  # 2/3
    p13 = 1  # 1/3
    o = 0
    m13 = -1  # -1/3
    m23 = -2  # -2/3
    m = -3
    m43 = -4  # -4/3
    mm = -6
    u = 50


class Inv(IntEnum):
    """Enum defining what happens when a particle is inverted.

    Possible Values
    ---------------
        Same     : particle = antiparticle, e.g. pi0.
        Barred   : antiparticle is denoted with a bar, e.g. proton, Lambda.
                   Note that the charge may or may not be part of the name, e.g. Lb0 vs neutrinos.
        ChargeInv: antiparticle is obtained with a change of charge, e.g. pi+ vs pi-.
    """

    Same = 0
    Barred = 1
    ChargeInv = 2


class Status(IntEnum):
    """
    The status of the particle, a one-letter code used by the PDG
    e.g. in the extended particle data table (PDT), see our .fwf files.
    The meanings are reproduced here for completeness,
    see also the Status_mapping dictionary in this module.
    RPP stands for the (PDG) Review of Particle Properties.

    Possible Values
    ---------------
        Common   : extended PDT code "R" - established particle
                   in RPP Summary Table in Particle Physics Booklet
                   (established quarks, gauge bosons, leptons, mesons and baryons,
                   except those in D below).
        Rare     : extended PDT code "D" - the particle is omitted from the
                   Summary Tables in Particle Physics Booklet, but not from the Review.
                   These entries are omitted only to save space even though they are well established.
        Unsure   : extended PDT code "S" - the particle is omitted from the
                   particle properties Summary Tables because it is not well established.
        Further  : extended PDT code "F" - special case "Further mesons", see RPP.
                   These states are in the RPP database but are poorly established
                   or observed by a single group and thus need confirmation.
        NotInPDT : an extra code (empty string "") we here use for non-standard
                   and exotic particles not in the PDT.
    """

    Common = 0
    Rare = 1
    Unsure = 2
    Further = 3
    NotInPDT = 4


# Mappings that allow the above classes to be produced from text mappings
Parity_mapping = {
    "+": Parity.p,
    "-": Parity.m,
    "?": Parity.u,
    "": Parity.u,
}
Charge_mapping = {
    "++": Charge.pp,
    "+4/3": Charge.p43,
    "+": Charge.p,
    "+2/3": Charge.p23,
    "+1/3": Charge.p13,
    "0": Charge.o,
    "-1/3": Charge.m13,
    "-2/3": Charge.m23,
    "-": Charge.m,
    "-4/3": Charge.m43,
    "--": Charge.mm,
    "?": Charge.u,
    "": Charge.u,
}

Inv_mapping = {"": Inv.Same, "F": Inv.Barred, "B": Inv.ChargeInv}
Status_mapping = {
    "R": Status.Common,
    "D": Status.Rare,
    "S": Status.Unsure,
    "F": Status.Further,
    "": Status.NotInPDT,
}

# Mappings that allow the above classes to be turned into text mappings
Parity_undo = {Parity.p: "+", Parity.m: "-", Parity.u: "None"}
Parity_prog = {Parity.p: "p", Parity.m: "m", Parity.u: "u"}

Charge_undo = {
    Charge.pp: "++",
    Charge.p43: "+4/3",
    Charge.p: "+",
    Charge.p23: "+2/3",
    Charge.p13: "+1/3",
    Charge.o: "0",
    Charge.m13: "-1/3",
    Charge.m23: "-2/3",
    Charge.m: "-",
    Charge.m43: "-4/3",
    Charge.mm: "--",
    Charge.u: "None",
}
Charge_prog = {
    Charge.pp: "pp",
    Charge.p43: "p43",
    Charge.p: "p",
    Charge.p23: "p23",
    Charge.p13: "p13",
    Charge.o: "0",
    Charge.m13: "m13",
    Charge.m23: "m23",
    Charge.m: "m",
    Charge.m43: "m43",
    Charge.mm: "mm",
    Charge.u: "u",
}



from __future__ import annotations

from pathlib import Path

import nox

nox.options.sessions = ["lint", "pylint", "tests"]


@nox.session
def lint(session: nox.Session) -> None:
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def pylint(session: nox.Session) -> None:
    """
    Run pylint.
    """

    session.install("pylint~=2.15.0")
    session.install("-e", ".[dev]")
    session.run("pylint", "src", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    session.install("-e", ".[test]")
    session.run(
        "pytest",
        *session.posargs,
    )


@nox.session
def build(session: nox.Session) -> None:
    """
    Build an SDist and wheel.
    """

    session.install("build", "twine", "check-wheel-contents")
    session.run("python", "-m", "build")
    session.run("twine", "check", "--strict", "dist/*")
    session.run(
        "check-wheel-contents", str(*Path("dist").glob("*.whl")), "--ignore=W002"
    )


@nox.session
def zipapp(session: nox.Session) -> None:
    tmpdir = session.create_tmp()

    # Build a distribution
    session.run(
        "python",
        "-m",
        "pip",
        "install",
        ".",
        "attrs",
        "hepunits",
        "importlib_resources",
        "deprecated",
        f"--target={tmpdir}",
    )

    # Build the zipapp out of the local directory
    outfile = Path("particle.pyz").resolve()
    session.chdir(tmpdir)
    session.run(
        "python",
        "-m",
        "zipapp",
        "--compress",
        "--python=/usr/bin/env python3",
        "--main=particle.__main__:main",
        f"--output={outfile}",
        ".",
    )

    # Quick test to verify it works
    session.chdir(str(outfile.parent))
    result = session.run("python", "particle.pyz", "search", "D0", silent=True)
    if "Name: D0" not in result:
        session.error(
            f"Expected valid result, was unable to run zipapp. Produced: {result}"
        )



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from ..geant import Geant3ID
from ..pdgid import PDGID
from .bimap import BiMap

Geant2PDGIDBiMap = BiMap(PDGID, Geant3ID)
Geant2PDGIDBiMap.__doc__ = """
Bi-bidirectional map between PDG and Geant3 IDs.

Examples
--------
>>> gid = Geant2PDGIDBiMap[PDGID(211)]
>>> gid
<Geant3ID: 8>

>>> pdgid = Geant2PDGIDBiMap[Geant3ID(8)]
>>> pdgid
<PDGID: 211>
"""



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Definitions of handy particle `PDGID` literals
==============================================

The PDGID literals provide aliases for most common particle PDGIDs, with easily recognisable names.
The aliases are instances of the PDGID class.

Typical use cases::

    >>> from particle.pdgid import literals as lid
    >>> lid.pi_plus
    <PDGID: 211>
    >>> lid.pi_plus.is_meson
    True
    >>> from particle.pdgid.literals import Lambda_b_0
    >>> Lambda_b_0
    <PDGID: 5122>
    >>> Lambda_b_0.has_bottom
    True

List of available/defined literals:

{0}
"""

from __future__ import annotations

from ..shared_literals import common_particles
from .pdgid import PDGID


def __dir__() -> list[str]:
    return list(common_particles)


for item in common_particles:
    locals()[item] = PDGID(common_particles[item])


__doc = "".join(
    f"  {item!s} = PDGID({common_particles[item]})\n" for item in common_particles
)
__doc__ = __doc__.format(__doc)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Class representing a PDG ID.

All methods of HepPID are implemented in a Pythonic version, see the functions module.
"""


from __future__ import annotations

from inspect import isfunction
from typing import TypeVar

from . import functions as _functions

# Collect all the user defined, non-hidden functions in the pdgid.functions module
_fnames = [
    fname
    for fname in dir(_functions)
    if not fname.startswith("_") and isfunction(getattr(_functions, fname))
]


Self = TypeVar("Self", bound="PDGID")


class PDGID(int):
    """
    Holds a PDG ID.

    Example
    -------
    >>> PDGID(11).is_lepton
    True
    """

    __slots__ = ()  # Keep PDGID a slots based class

    def __repr__(self) -> str:
        is_valid_str = "" if _functions.is_valid(self) else " (is_valid==False)"
        return f"<{self.__class__.__name__}: {int(self):d}{is_valid_str}>"

    def __str__(self) -> str:
        return repr(self)

    def __neg__(self: Self) -> Self:
        return self.__class__(-int(self))

    __invert__ = __neg__

    def info(self) -> str:
        """
        Print all PDGID properties one per line, for easy inspection.
        """
        return "".join(f"{item:14} {getattr(self, item)}\n" for item in _fnames)

    A = property(_functions.A, doc=_functions.A.__doc__)
    J = property(_functions.J, doc=_functions.J.__doc__)
    L = property(_functions.L, doc=_functions.L.__doc__)
    S = property(_functions.S, doc=_functions.S.__doc__)
    Z = property(_functions.Z, doc=_functions.Z.__doc__)
    abspid = property(_functions.abspid, doc=_functions.abspid.__doc__)
    charge = property(_functions.charge, doc=_functions.charge.__doc__)
    has_bottom = property(_functions.has_bottom, doc=_functions.has_bottom.__doc__)
    has_charm = property(_functions.has_charm, doc=_functions.has_charm.__doc__)
    has_down = property(_functions.has_down, doc=_functions.has_down.__doc__)
    has_fundamental_anti = property(
        _functions.has_fundamental_anti, doc=_functions.has_fundamental_anti.__doc__
    )
    has_strange = property(_functions.has_strange, doc=_functions.has_strange.__doc__)
    has_top = property(_functions.has_top, doc=_functions.has_top.__doc__)
    has_up = property(_functions.has_up, doc=_functions.has_up.__doc__)
    is_Qball = property(_functions.is_Qball, doc=_functions.is_Qball.__doc__)
    is_Rhadron = property(_functions.is_Rhadron, doc=_functions.is_Rhadron.__doc__)
    is_SUSY = property(_functions.is_SUSY, doc=_functions.is_SUSY.__doc__)
    is_baryon = property(_functions.is_baryon, doc=_functions.is_baryon.__doc__)
    is_diquark = property(_functions.is_diquark, doc=_functions.is_diquark.__doc__)
    is_dyon = property(_functions.is_dyon, doc=_functions.is_dyon.__doc__)
    is_excited_quark_or_lepton = property(
        _functions.is_excited_quark_or_lepton,
        doc=_functions.is_excited_quark_or_lepton.__doc__,
    )
    is_gauge_boson_or_higgs = property(
        _functions.is_gauge_boson_or_higgs,
        doc=_functions.is_gauge_boson_or_higgs.__doc__,
    )
    is_generator_specific = property(
        _functions.is_generator_specific, doc=_functions.is_generator_specific.__doc__
    )
    is_hadron = property(_functions.is_hadron, doc=_functions.is_hadron.__doc__)
    is_lepton = property(_functions.is_lepton, doc=_functions.is_lepton.__doc__)
    is_meson = property(_functions.is_meson, doc=_functions.is_meson.__doc__)
    is_nucleus = property(_functions.is_nucleus, doc=_functions.is_nucleus.__doc__)
    is_pentaquark = property(
        _functions.is_pentaquark, doc=_functions.is_pentaquark.__doc__
    )
    is_quark = property(_functions.is_quark, doc=_functions.is_quark.__doc__)
    is_sm_gauge_boson_or_higgs = property(
        _functions.is_sm_gauge_boson_or_higgs,
        doc=_functions.is_sm_gauge_boson_or_higgs.__doc__,
    )
    is_sm_lepton = property(
        _functions.is_sm_lepton, doc=_functions.is_sm_lepton.__doc__
    )
    is_sm_quark = property(_functions.is_sm_quark, doc=_functions.is_sm_quark.__doc__)
    is_special_particle = property(
        _functions.is_special_particle, doc=_functions.is_special_particle.__doc__
    )
    is_technicolor = property(
        _functions.is_technicolor, doc=_functions.is_technicolor.__doc__
    )
    is_valid = property(_functions.is_valid, doc=_functions.is_valid.__doc__)
    j_spin = property(_functions.j_spin, doc=_functions.j_spin.__doc__)
    l_spin = property(_functions.l_spin, doc=_functions.l_spin.__doc__)
    s_spin = property(_functions.s_spin, doc=_functions.s_spin.__doc__)
    three_charge = property(
        _functions.three_charge, doc=_functions.three_charge.__doc__
    )


# Verify the PDGID class has all relevant functions defined in the pdgid.functions module
for _n in _fnames:
    assert _n in dir(
        PDGID
    ), f"{_n} missing from PDGID class! Update the list in pdgid.py"



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import math
import re
import unicodedata
from html.entities import name2codepoint


def programmatic_name(name: str) -> str:
    "Return a name safe to use as a variable name."
    name = re.sub("0$", "_0", name)
    # Deal first with antiparticles of sparticles, e.g. "~d(R)~" antiparticle of "~d(R)"
    name = re.sub("^~", "tilde_", name)
    # The remaining "~" now always means it's an antiparticle
    name = name if "~" not in name else "".join(name.split("~")) + "_bar"
    name = (
        name.replace(")(", "_")
        .replace("(", "_")
        .replace(")", "")
        .replace("*", "st")
        .replace("'", "p")
        .replace("::", "_")
        .replace("/", "")
        .replace("--", "_mm")
        .replace("++", "_pp")
        .replace("-", "_minus")
        .replace("+", "_plus")
    )
    # Strip off the ugly "_" at beginning of a name, such as when dealing with name="(bs)(0)""
    return name.lstrip("_")


def str_with_unc(value: float, upper: float | None, lower: float | None = None) -> str:
    """
    Utility to print out an uncertainty with different or
    identical upper/lower bounds. Nicely formats numbers using PDG rule.
    """

    # If no errors are available, simply return the value alone
    if upper is None:
        return str(value)

    # If no lower passed, make them the same
    if lower is None:
        lower = upper

    # Uncertainties are always positive
    upper = abs(upper)
    lower = abs(lower)

    error = min(upper, lower)

    if error == 0:
        return str(value)

    value_digits = int(math.floor(math.log10(value)))
    error_digits = int(math.floor(math.log10(error) - math.log10(2.5)))
    # This is split based on the value being larger than 1000000 or smaller than 0.001 - scientific notation split

    # This is normal notation
    if -3 < value_digits < 6:
        if error_digits < 0:
            fsv = fse = f".{-error_digits}f"
        else:
            fsv = fse = ".0f"

    # This is scientific notation - a little odd, but better than the other options.
    else:
        fsv = f".{abs(error_digits - value_digits)}e"
        pure_error_digits = int(math.floor(math.log10(error)))

        fse = ".0e" if error_digits == pure_error_digits else ".1e"

    # Now, print values based on upper=lower being true or not (even if they print the same)
    if upper != lower:
        return f"{value:{fsv}} + {upper:{fse}} - {lower:{fse}}"

    return f"{value:{fsv}} ± {upper:{fse}}"


# List of greek letter names as used in Unicode (see unicodedata package)
_list_name_greek_letters = [
    "Alpha",
    "Beta",
    "Chi",
    "Delta",
    "Epsilon",
    "Eta",
    "Gamma",
    "Iota",
    "Kappa",
    "Lambda",
    "Mu",
    "Nu",
    "Omega",
    "Omicron",
    "Phi",
    "Pi",
    "Psi",
    "Rho",
    "Sigma",
    "Tau",
    "Theta",
    "Upsilon",
    "Xi",
    "Zeta",
]
_list_name_greek_letters += [item.lower() for item in _list_name_greek_letters]


def greek_letter_name_to_unicode(letter: str) -> str:
    """
    Return a greek letter name as a Unicode character,
    the same as the input if no match is found.

    Examples
    --------
    Lamda -> Λ    (Unicodedata library uses "lamda" for "lambda" :S!)
    Omega -> Ω
    omega -> ω
    """
    case = "SMALL" if letter == letter.lower() else "CAPITAL"
    name = letter.upper()

    try:
        return unicodedata.lookup(f"GREEK {case} LETTER {name}")
    except KeyError:  # Unicodedata library uses "lamda" for "lambda", so that's an obvious miss
        return letter


def latex_name_unicode(name: str) -> str:
    r"""
    Convert in particle names in LaTeX all greek letters by their unicode.

    Examples
    --------
    >>> from particle import Particle
    >>> n = Particle.from_pdgid(3124).latex_name
    >>> print(n)
    \Lambda(1520)
    >>> latex_name_unicode(n)
    'Λ(1520)'
    """
    # Make sure "Lambda" and "lambda" are naturally deal with given that the
    # unicodedata library uses "lamda" for "lambda" :S!
    if "ambda" in name:
        name = name.replace("ambda", "amda")
    for gl in _list_name_greek_letters:
        name = name.replace(rf"\{gl}", greek_letter_name_to_unicode(gl))
    return name


def latex_to_html_name(name: str) -> str:
    """Conversion of particle names from LaTeX to HTML."""
    name = re.sub(r"\^\{(.*?)\}", r"<SUP>\1</SUP>", name)
    name = re.sub(r"\_\{(.*?)\}", r"<SUB>\1</SUB>", name)
    name = re.sub(r"\\prime(.*?)", r"&#8242;", name)
    name = re.sub(r"\\mathrm\{(.*?)\}", r"\1", name)
    name = re.sub(r"\\left\[(.*?)\\right\]", r"[\1] ", name)
    for gl in _list_name_greek_letters:
        # Special formatting since for example
        # f"{hex(html.entities.name2codepoint['Delta'])}" gives '0x394' whereas HTML needs 'x0394',
        # as in '&#x0394;', equivalent to '&Delta;'
        name = name.replace(rf"\{gl}", f"&#x{name2codepoint[gl]:04x};")
    name = re.sub(r"\\tilde\{(.*?)\}", r"\1&#771;", name)
    name = re.sub(r"\\overline\{(.*?)\}", r"\1&#773;", name)
    name = re.sub(r"\\bar\{(.*?)\}", r"\1&#773;", name)
    return name



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import sys

if sys.version_info < (3, 9):
    import importlib_resources as resources
else:
    from importlib import resources

__all__ = ["basepath"]


basepath = resources.files(__name__)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from .enums import Charge, Inv, Parity, SpinType, Status
from .kinematics import lifetime_to_width, width_to_lifetime
from .particle import InvalidParticle, Particle, ParticleNotFound
from .utilities import latex_name_unicode, latex_to_html_name, programmatic_name

__all__ = (
    "Charge",
    "Inv",
    "Parity",
    "SpinType",
    "Status",
    "lifetime_to_width",
    "width_to_lifetime",
    "InvalidParticle",
    "Particle",
    "ParticleNotFound",
    "latex_name_unicode",
    "latex_to_html_name",
    "programmatic_name",
)


def __dir__() -> tuple[str, ...]:
    return __all__



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from ..corsika import Corsika7ID
from ..pdgid import PDGID
from .bimap import BiMap

Corsika72PDGIDBiMap = BiMap(PDGID, Corsika7ID)
Corsika72PDGIDBiMap.__doc__ = """
Bi-bidirectional map between PDG and Corsika7 IDs.

Examples
--------
>>> cid = Corsika72PDGIDBiMap[PDGID(-13)]
>>> cid
<Corsika7ID: 13>

>>> cid = Corsika72PDGIDBiMap[Corsika7ID(5)]
>>> cid
<PDGID: -13>
"""



from __future__ import annotations

import pytest

from particle.corsika import Corsika7ID
from particle.exceptions import MatchingIDNotFound
from particle.pdgid import PDGID


def test_class_string_representations():
    pid = Corsika7ID(1)
    assert pid == 1
    assert pid.__str__() == "<Corsika7ID: 1>"


def test_class_return_type():
    assert isinstance(Corsika7ID(3), Corsika7ID)


def test_from_pdgid():
    assert Corsika7ID.from_pdgid(-13) == 5

    assert Corsika7ID.from_pdgid(PDGID(-13)) == 5
    assert Corsika7ID.from_pdgid(PDGID(13)) == Corsika7ID(6)


def test_from_pdgid_non_matching():
    with pytest.raises(MatchingIDNotFound):
        Corsika7ID.from_pdgid(55)


def test_to_pdgid():
    cid = Corsika7ID(5)
    assert cid.to_pdgid() == -13
    assert cid.to_pdgid() == PDGID(-13)


def test_is_particle():
    cid = Corsika7ID(1)
    assert cid.is_particle()
    cid = Corsika7ID(75)
    assert not cid.is_particle()


def test_from_particle_description():
    cid, is_mother = Corsika7ID.from_particle_description(-6001)
    assert is_mother
    assert cid.is_particle()
    cid, is_mother = Corsika7ID.from_particle_description(75001)
    assert not is_mother
    assert not cid.is_particle()
